﻿// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameActor.h"
#include "IwGameScene.h"
#include "IwGameCamera.h"


//
//
//
// CIwGameActor implementation
//
//
//

CIwGameActor::~CIwGameActor()
{
	if (Visual != NULL)
		Scene->getSpriteManager()->removeSprite(Visual);

	if (VisualAnimManager != NULL)
	{
		delete VisualAnimManager;
		VisualAnimManager = NULL;
	}
}

void CIwGameActor::setCollisionRect(CIwRect& rect)
{
	CollisionRect = rect;
	int dx = rect.w >> 1;
	int dy = rect.y >> 1;
	CollisionSize = dx * dx + dy * dy;
}

void CIwGameActor::Reset()
{
	Visual = NULL;
	VisualAnimManager = NULL;
	IsActive = true;
	IsVisible = true;
	IsCollidable = false;
	WrapPosition = false;
	Position.x = 0;
	Position.y = 0;
	OriginalPosition.x = 0;
	OriginalPosition.y = 0;
	Scale = 1.0f;
	OriginalAngle = 0;
	Angle = 0;
	Velocity.x = 0;
	Velocity.y = 0;
	VelocityDamping.x = 1.0f;
	VelocityDamping.y = 1.0f;
	AngularVelocity = 0;
	AngularVelocityDamping = 1.0f;
	CollisionSize = 0;
	Colour.r = 0xff;
	Colour.g = 0xff;
	Colour.b = 0xff;
	Colour.a = 0xff;
}

bool CIwGameActor::Update(float dt)
{
	if (!IsActive)
		return true;

	// Update velocity
	PreviousPosition = Position;
	Position += Velocity * dt;
	Velocity.x *= VelocityDamping.x;
	Velocity.y *= VelocityDamping.y;
	PreviousAngle = Angle;
	Angle += AngularVelocity * dt;
	AngularVelocity *= AngularVelocityDamping;

	// Wrap actor position with scene extents (if enabled)
	CIwRect extents = Scene->getExtents();
	if (WrapPosition)
	{
		if (Position.x < extents.x)
			Position.x += extents.w;
		else
		if (Position.x >= (extents.x + extents.w))
			Position.x -= extents.w;
		if (Position.y < extents.y)
			Position.y += extents.h;
		else
		if (Position.y >= (extents.y + extents.h))
			Position.y -= extents.h;
	}

	// Update visual animation
	if (VisualAnimManager != NULL)
		VisualAnimManager->Update(dt);

	// If this actor is collidable then add it to the scenes potential collidables list
	if (IsCollidable)
		Scene->addCollideable(this);

	return true;
}

bool CIwGameActor::UpdateVisual()
{
	if (!IsVisible || !IsActive || Visual == NULL)
		return false;

	CIwFVec2 cam_pos = Scene->getCamera()->getPosition();
	// Transfor actor properties to the visual
	Visual->setPosAngScale(	(int)(Position.x - cam_pos.x), 
							(int)(Position.y - cam_pos.y), 
							(iwangle)((Angle * IW_ANGLE_2PI) / 360.0f), (iwfixed)(Scale * IW_GEOM_ONE));
	Visual->setColour(Colour);

	return true;
}

bool CIwGameActor::CheckIfTapped()
{
	return false;
}

bool CIwGameActor::CheckCollision(CIwGameActor* other)
{
	// This is just about as simple as colision checks can go. We check to see if the distance between the centre of both objects is less than the sum of the radii of both
	// Note that we do not use a proper distance between two points check because that involves expensive square root calculations, but this works just the same
	// The only thing to be careful of is if your coordinate system is huge and your numbers have a large range, just ensure that the term for d = dx * dx + dy * dy does not fall 
	// out of integer range
	CIwFVec2 other_pos = other->getPosition();
	int		other_collision_size = other->getCollisionSize();
	int		dx = (int)(Position.x - other_pos.x);
	int		dy = (int)(Position.y - other_pos.y);
	int		d_max = (CollisionSize >> 1) + (other_collision_size >> 1);
	int		d = dx * dx + dy * dy;

	d_max *= d_max;
	if (d <= d_max)
		return true;

	return false;
}

//
//
//
// CIwGameActorImage implementation
//
//
//

bool CIwGameActorImage::Init(CIwGameScene* scene, CIwGameImage* image, CIwGameAnimImage* anim, int width, int height)
{
	// Reset the actor
	CIwGameActor::Reset();

	// Set the scene
	Scene = scene;

	// Create sprite
	if (image != NULL)
	{
		CIwGameBitmapSprite* sprite = new CIwGameBitmapSprite();
		if (sprite == NULL)
			return false;

		// Set sprite image
		sprite->setImage(image);
		sprite->setDestSize(width, height);

		// Set sprite as visual
		Visual = sprite;

		// Add sprite to the sprite manager so it can be managed and drawn
		Scene->getSpriteManager()->addSprite(sprite);
	}

	// Create an animation manager and add the animation to it
	if (anim != NULL)
	{
		VisualAnimManager = new CIwGameAnimManager();
		VisualAnimManager->setUpdateAll(false);
		if (VisualAnimManager == NULL)
			return false;
		VisualAnimManager->addAnimation(anim);

		// Set the first animation in the animation manager as the current animation 
		VisualAnimManager->setCurrentAnimation(0);
	}

	return true;
}

bool CIwGameActorImage::UpdateVisual()
{
	if (CIwGameActor::UpdateVisual())
	{
		// Update the sprite
		if (VisualAnimManager != NULL)
		{
			// Get the animations current image frame data and copy it to the bitmapped sprite
			CIwGameAnimImage* anim = (CIwGameAnimImage*)VisualAnimManager->getAnimation(0);
			if (Visual != NULL)
				((CIwGameBitmapSprite*)Visual)->setSrcRect(anim->getCurrentFrameData());
		}

		return true;
	}

	return false;
}

