// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameImage.h"
#include "IwGameFile.h"

//
//
// JPEG utility
//
//
extern "C" 
{
#include "jpeglib.h"
}

typedef struct 
{
	jpeg_source_mgr	pub;
	char*			buf;
	char			buf_term[2];
	long			buf_size;
	long			pos;
	bool			read_started;
} buf_source_mgr;

void init_source_from_buf(j_decompress_ptr cinfo)
{
	buf_source_mgr* src = (buf_source_mgr*) cinfo->src;
	src->read_started = true;
}

void skip_input_data_from_buf(j_decompress_ptr cinfo, long nbytes)
{
	buf_source_mgr* src = (buf_source_mgr*)cinfo->src;
	if (nbytes > 0)
	{
		src->pub.next_input_byte += (size_t) nbytes;
		src->pub.bytes_in_buffer -= (size_t) nbytes;
	}
}

boolean fill_input_buffer_from_buf(j_decompress_ptr cinfo)
{
	buf_source_mgr* src = (buf_source_mgr*)cinfo->src;

	if (src->pos == src->buf_size)
	{
		src->buf_term[0] = (JOCTET) 0xFF;
		src->buf_term[1] = (JOCTET) JPEG_EOI;
		src->pub.next_input_byte = (JOCTET*)src->buf_term;
		src->pub.bytes_in_buffer = 2;
		src->read_started = false;
		return TRUE;
	}

	src->pub.next_input_byte = (JOCTET*)src->buf;
	src->pub.bytes_in_buffer = src->buf_size;
	src->pos = src->buf_size;
	src->read_started = false;

	return TRUE;
}

void term_source_from_buf(j_decompress_ptr cinfo)
{
}

void jpeg_buf_src (j_decompress_ptr cinfo, char* buf,long size)
{
	buf_source_mgr* src = (buf_source_mgr*) cinfo->src;
	if (cinfo->src == NULL)
	{
		cinfo->src = (struct jpeg_source_mgr *)
		(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT, sizeof(buf_source_mgr));
		src = (buf_source_mgr*)cinfo->src;
	}

	src = (buf_source_mgr*) cinfo->src;
	src->pub.init_source = init_source_from_buf;
	src->pub.fill_input_buffer = fill_input_buffer_from_buf;
	src->pub.skip_input_data = skip_input_data_from_buf;
	src->pub.resync_to_restart = jpeg_resync_to_restart;
	src->pub.term_source = term_source_from_buf;
	src->pub.bytes_in_buffer = 0;
	src->pub.next_input_byte = (JOCTET*)NULL;

	src->buf = buf;
	src->read_started = false;
	src->buf_size = size;
	src->pos = 0;
}

bool IsJPEG(const char *jpeg_data, int length)
{
	const char pJPEGSignature[] = { 0xFF, 0xD8, 0xFF, 0xE0, 0, 0, 0x4A, 0x46, 0x49, 0x46, 0x00 };

	if (length > (int)sizeof(pJPEGSignature))
	{
		if (!memcmp(jpeg_data, pJPEGSignature, 4) && !memcmp(jpeg_data + 6, pJPEGSignature + 6, 5))
		{
			return true;
		}
	}

	return false;
}


//
//
// CIwGameImage implementation
//
//
bool CIwGameImage::Load()
{
	// If already loaded return true
	if (Image2D != NULL)
		return true;

	// Create image from the resource group
	IwGetResManager()->SetCurrentGroup(ResourceGroup);
	Image2D = Iw2DCreateImageResource(NameHash);
	if (Image2D != NULL)
	{
		Width = Image2D->GetWidth();
		Height = Image2D->GetHeight();
		State = CIwGameImage_State_Loaded;
		return true;
	}

	return false;
}

bool CIwGameImage::Init(const char* name, void* memory_file, int memory_file_size, bool is_jpeg)
{
	CIwGameFile file;
	if (file.Open(memory_file, memory_file_size))
	{
		if (is_jpeg)
		{
			if (!DecompressJPEG((char*)memory_file, memory_file_size))
				return false;
		}
		else
		{
			CIwImage image;
			image.ReadFile(file.getFileHandle());
			Image2D = Iw2DCreateImage(image);
			Width = Image2D->GetWidth();
			Height = Image2D->GetHeight();
		}
	}

	State = CIwGameImage_State_Loaded;

	return true;
}

bool CIwGameImage::DecompressJPEG(char* jpeg_data, int jpeg_data_size)
{
	// Decompress the JPEG data
	int			length = jpeg_data_size;
	struct		jpeg_error_mgr jerr;
	struct		jpeg_decompress_struct cinfo;

	cinfo.err = jpeg_std_error(&jerr);

	jpeg_create_decompress	(&cinfo);
	jpeg_buf_src(&cinfo,(char*)jpeg_data, length);
	jpeg_read_header	(&cinfo, TRUE);
	cinfo.out_color_space = JCS_RGB;
	jpeg_start_decompress	(&cinfo);
	
	int newlen = cinfo.image_width * cinfo.image_height * 3;

	unsigned char * data = (unsigned char*)s3eMalloc(newlen);
	if (data == NULL)
	{
		jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);
		return false;
	}
	unsigned char * linha = data;
   
	while (cinfo.output_scanline < cinfo.output_height)
	{
		linha = data + 3 * cinfo.image_width * cinfo.output_scanline;
		jpeg_read_scanlines(&cinfo,&linha,1);
	}
    
	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

	Width = cinfo.output_width;
	Height = cinfo.output_height;

//	int stride = cinfo.output_components;

	// Create CIw2DImage from pixel data
	CIwImage image;
	image.SetFormat(CIwImage::RGB_888);
	image.SetWidth(Width);
	image.SetHeight(Height);
	image.SetOwnedBuffers((uint8*)data, 0);
	Image2D = Iw2DCreateImage(image);

//	s3eFree(data);

	return true;
}


//
//
// CIwGameImageManager implementation
//
//

void CIwGameImageManager::addImage(const char* resource_name, CIwResGroup* resource_group, bool preload)
{
	CIwGameImage* image = new CIwGameImage();
	image->Init(resource_name, resource_group);
	addImage(image);
	if (preload)
		image->Load();
}

void CIwGameImageManager::addImage(CIwGameImage *image)
{
	// Add an image to the manager
	Images.push_back(image);
	image->setParent(this);
}

void CIwGameImageManager::removeImage(CIwGameImage* image)
{
	removeImage(image->getNameHash());
}

void CIwGameImageManager::removeImage(unsigned int name_hash)
{
	// Renove named image from the manager
	for (_Iterator it = Images.begin(); it != Images.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
		{
			delete *it;
			Images.erase(it);
			break;
		}
	}
}

CIwGameImage* CIwGameImageManager::findImage(unsigned int name_hash)
{
	// Find and eturn named image
	for (_Iterator it = Images.begin(); it != Images.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
			return *it;
	}

	return NULL;
}

CIwGameImage* CIwGameImageManager::findImage(const char* name)
{
	return findImage(IwHashString(name));
}

void CIwGameImageManager::clearImages()
{
	// Remove and delete all images from the manager
	for (_Iterator it = Images.begin(); it != Images.end(); ++it)
	{
		delete *it;
	}
	Images.clear();
}
